import ctypes
import sys
from PyQt5.QtWidgets import QApplication
from molmag_ac_gui.ac_gui import *

if __name__ == '__main__':
    
    myappid = 'AC Processing v1.0'
    ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)
    
    app = QApplication(sys.argv)
    w = ACGui()
    sys.exit(app.exec_())